/*
** gif2lss
**
**  V0.1 read gif images with giflib
**  V0.2 output first gif image
**  V0.3 encode lines with LSS runlength encoding
**  V0.4 produce useable binary output
**
** Copyright (C) 2001 Herbert Ptzl
** herbert@13thfloor.at
**
**  if you think you should contact me, ask yourself again,
**  and, if your decision still stands, feel free to do so.
**
** ------------------
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
*/

#define VERSION "V0.4"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "gif_lib/gif_lib.h"

#define	MAX_WIDTH	640

#define	MIN(a,b)	(((a)<(b))?(a):(b))
#define MAX(a,b)	(((a)>(b))?(a):(b))

#define NSWAP(a)	((((a)>>4)&0xf)|(((a)<<4)&0xf0))	



static	FILE *lss_stream;


static 	inline
void	LSSOutByte(
	int value)
{
	fputc((value & 0xff), lss_stream);
}

void	LSSOutNybble(
	int value)
{
	static	unsigned nytmp = 0;
	static	char odd = 0;
	register char val = (value & 0xf);
	
	if (value<0) {
	    if (odd) {
	    	LSSOutByte((nytmp << 0));
	    	odd = 0;
	    }
	} else {
	    if (odd) {
	    	LSSOutByte((nytmp << 0) | (val << 4));
	    } else nytmp = val;
	    odd = ~odd;
	}
}	


void	LSSDumpLine(
	const unsigned char *data, 
	unsigned width)
{
	int pixel;
	unsigned index, last;
	unsigned count;
	
	last=0; count=0;
	for (index=0; index<=width; index++) {
	    pixel = (index==width)?-1:(data[index] & 0xf);
	    if (pixel == last) count++;
	    if ((count == (16+255)) || (pixel != last)) {
	    	if (count) {
		    LSSOutNybble(last);
		    if (count < 16) {
		    	LSSOutNybble(count);
		    } else {
		    	LSSOutNybble(0);
		    	LSSOutNybble((count-16) & 0xf);
		    	LSSOutNybble((count-16) >> 4);
		    }
		    count=0;
		}
	    }
	    if (pixel != last) {
	    	LSSOutNybble(pixel);
		last = pixel;
	    }
	}
}

void	LSSDumpHeader(
	unsigned width, 
	unsigned height, 
	GifColorType *cmap)
{
	unsigned index;
	
	fputc(0x3d, lss_stream);
	fputc(0xf3, lss_stream);
	fputc(0x13, lss_stream);
	fputc(0x14, lss_stream);
	fputc((width & 0xff), lss_stream);
	fputc((width >> 8), lss_stream);
	fputc((height & 0xff), lss_stream);
	fputc((height >> 8), lss_stream);

	for (index=0; index<16; index++) {
	    fputc(cmap[index].Red >> 2, lss_stream);
	    fputc(cmap[index].Green >> 2, lss_stream);
	    fputc(cmap[index].Blue >> 2, lss_stream);
	}
}


int	main(int argc, char *argv[])
{
	GifFileType *gf;
	SavedImage *im;
	GifColorType cmap[16];
	GifPixelType *data;
	
	unsigned index, line, ncols;
	unsigned width, height;

	lss_stream = stdout;
	
	if (argc <= 1) {
	    fprintf(stderr, 
	    	"This is %s " VERSION "\n"
		"%s <source-gif> [> <destination> ]\n",
		argv[0],argv[0]);
	    exit(-1);
	}
	gf = DGifOpenFileName(argv[1]);	
	if (!gf) exit(-1);
	if (DGifSlurp(gf)==GIF_OK) {
	    width = MIN(gf->SWidth, MAX_WIDTH);
	    height = gf->SHeight;
	    fprintf(stderr, "IMAGE: %dx%d\n", width, height);
	    if (gf->SColorMap) {
	        fprintf(stderr, "GCMAP: %d (%d)\n", 
		    gf->SColorMap->ColorCount, 
		    gf->SColorMap->BitsPerPixel);
		ncols = MIN(16, gf->SColorMap->ColorCount);
		for (index=0; index<ncols; index++) {
		    cmap[index] = gf->SColorMap->Colors[index];
		}
	    }
	    im = &(gf->SavedImages[0]);	    
	    width = MIN(im->ImageDesc.Width, MAX_WIDTH);
	    height = im->ImageDesc.Height;
	    fprintf(stderr, "IDESC: %dx%d\n", width, height);
	    if (im->ImageDesc.ColorMap) {
	        fprintf(stderr, "ICMAP: %d (%d)\n", 
		    im->ImageDesc.ColorMap->ColorCount, 
		    im->ImageDesc.ColorMap->BitsPerPixel);
		ncols = MIN(16, im->ImageDesc.ColorMap->ColorCount);
		for (index=0; index<ncols; index++) {
		    cmap[index] = im->ImageDesc.ColorMap->Colors[index];
		}
	    }

	    LSSDumpHeader(width, height, cmap);
	    data = im->RasterBits;
	    for (line=0; line<height; line++) {
	    	LSSDumpLine(data, width);
		data += im->ImageDesc.Width;
	    }
	} else {
	    PrintGifError();
	}
	DGifCloseFile(gf);
	return 0;
}

